import { collection } from '@metromerce/firestore'
import { optionRequireRefetch } from './helper'

import namer from './namer'
export const REQUEST = (options) => namer(options) + '_COLLECTION_REQUEST'
export const RESPONSE_SUCCESS = (options) => namer(options) + '_COLLECTION_RESPONSE_SUCCESS'
export const RESPONSE_FAIL = (options) => namer(options) + '_COLLECTION_RESPONSE_FAIL'
export const SUBSCRIBE = (options) => namer(options) + '_COLLECTION_SUBSCRIBE'
export const UNSUBSCRIBE = (options) => namer(options) + '_COLLECTION_UNSUBSCRIBE'


function fetchIfNeeded(options) {
  return (dispatch, getState) => {
    dispatch(unsubscribeIfNeeded(options))
    const db = getState().getIn(['firestore','getFsState'])(options)
    if(shouldFetch(db, options))
      return dispatch(fetch(db, options))
    else {
      if(options.callback && db.pages[options.pageIndex])
          options.callback (Object.values(db.pages[options.pageIndex].items))
      return Promise.resolve()
    }
  }
}

function shouldFetch(db, options) {
  if(options.onSnapshot && options.pageLength)
    throw Error("'pageLength' with 'onSnapshot' options does not support")
  if(options.onSnapshot)
    return true // always fetch if subscribing regardless of the cache exist
  if(db && db.lastOption && optionRequireRefetch(db.lastOption, options))
    return true // option change requires refetch
  if(db && db.pages) {
    const page = db.pages[options.pageIndex]
    if(page && page.isFetching)
      return false
    if(!page || page.error)
      return true
  }
  return !db
}

function fetch(db, options) {
  return (dispatch, getState) => {
    dispatch(request(options))

    var queryParams = {...options}
    if(options.pageIndex > 0) {
      var prevPageIndex = options.pageIndex-1
      var _lastSuccessPageIndex = db.lastSuccessOption ? db.lastSuccessOption.pageIndex : -1
      if(prevPageIndex > _lastSuccessPageIndex)
        return dispatch(responseFail(options, Error('invalid page index:' + options.pageIndex)))
      if(options.orderBy) {
        var prevPage = db.pages[prevPageIndex]
        if(prevPage && prevPage.fetched && prevPage.itemIds.length > 0)
          queryParams.startAfter = prevPage.items[prevPage.itemIds[prevPage.itemIds.length-1]][Object.keys(options.orderBy)[0]]
        else
          return dispatch(responseFail(options, Error('no previous page')))
      }
    }

    if(queryParams.onSnapshot) {
      dispatch(unsubscribeIfNeeded(options))
      const unsubscribe = collection.onSnapshot(queryParams)
      dispatch(subscribe(options, unsubscribe))
    }
    else {
      collection.get(queryParams)
        .then(result => {
          dispatch(responseSucesss(options, result))
        })
        .catch(error => {
          dispatch(responseFail(options, error))
        })
    }
  }
}

function request(options) {
  return { type:REQUEST(options),options }
}

function responseSucesss(options, result) {
  return dispatch => {
    dispatch({ type:RESPONSE_SUCCESS(options), options, result })
    if(options.callback)
      options.callback(result)
  }
}

function responseFail(options, error) {
  return { type:RESPONSE_FAIL(options), options, error }
}

function subscribe(options, unsubscribe ) {
  return { type:SUBSCRIBE(options), options, unsubscribe }
}

function unsubscribeIfNeeded(options) {
  return (dispatch, getState) => {
    //const state = getState().firestore
    const db = getState().getIn(['firestore','getFsState'])(options)
    if(db && db.unsubscribe)
      dispatch({ type:UNSUBSCRIBE(options), options })
    else
      Promise.resolve()
  }
}

export default () => {
  return {
    fetchIfNeeded: (options) => {
      return fetchIfNeeded(options)
    },
    onSnapshot: (options, result) => {
      return responseSucesss(options, result)
    },
    unsubscribeIfNeeded: (options) => {
      return unsubscribeIfNeeded(options)
    }
  }
}
